# Protocol for communicating with leapd

The Leap Motion service provides a locally running WebSocket server listening to port 6347. This server provides tracking data as JSON.

## Typical sequence of events on connecting

* Connects to `/v4.json`
* Server resonds with `{version: 4}`
* Client sends `{enableGestures:true}`
* Client sends `{focused:true}`
* Servers sends frames (one message per frame)
* Client loses the OSLeap Motion control focus and sends `{focused:false}`
* Servers stops sending frames

## Negotiating protocol

The client requests a specific version of the protocol by requesting a specific path. The server responds with a message indicating what protocol it can respond with. It is assumed that any client can currently speak and protocol lower than what it is requesting. Support for a minimum supported protocol will be added later. The latest version of the protocol is: `v4.json`.

## Version 1

The first version of the protocol is available at `/` or `/v1.json`.

### Sending

#### Enable gestures

* `{ enableGestures: {true, false} }`

You can send a message at any point to enable or disable gesture support.

### Receiving

Each frame of data from the WebSocket server contains JSON defining a frame. The attributes of a frame in the JSON message are similar, but not identical to those of a Frame object obtained through the native library. They include:

```
"id": float
"timestamp": integer

"hands": array of Hand objects
   "direction": array of floats (vector)
   "id": integer
   "palmNormal": array of floats (vector)
   "palmPosition": array of floats (vector)
   "palmVelocity": array of floats (vector)
   "r": nested arrays of floats (matrix)
   "s": float
   "sphereCenter": array of floats (vector)
   "sphereRadius": float
   "t": array of floats (vector)

"interactionBox": object
   "center": array of floats (vector)
   "size": array of floats (vector)

"pointables": array of Pointable objects
   "direction": array of floats (vector)
   "handId": integer
   "id": integer
   "length": float
   "stabilizedTipPosition": array of floats (vector)
   "tipPosition":  array of floats (vector)
   "tipVelocity":  array of floats (vector)
   "tool": boolean (true or false)
   "touchDistance": float
   "touchZone": string - one of "none", "hovering", "touching"

"gestures": array of Gesture objects
    "center": array of floats (vector)
    "duration": integer microseconds
    "handIds": array of integers
    "id": integer
    "normal": array of floats
    "pointableIds": array
    "progress": float,
    "radius": float,
    "state": string - one of "start", "update", "stop"
    "type": string - one of "circle", "swipe", "keyTap", "screenTap"

"r": nested arrays of floats (matrix)
"s": float
"t": array of floats (vector)
```

#### Motion factors

The motion factors, r, s, t, attached to Hand and Frame objects are snapshots of the motion occuring across frames. These factors must be combined with those of a previous frame to derive the relative motion.

* r -- a 3x3 rotation matrix
* s -- a scale factor
* t -- a 3-element translation vector

#### Rotation factor

The matrix expressing the relative rotation between two frames can be calculated by multiplying the r matrix from the current frame by the inverse of the r matrix of the starting frame.

#### Scale factor

The relative scale factor between two frames can be calculated by subtracting the s value from the starting frame from the current s value and taking the natural logarithm of the result.

#### Translation factor

The relative translation factor between two frames can be calculated by subtracting the t vector from the starting frame from the current t factor.

## Version 2

### Changes

This version introduced *heartbeats*. Heartbeats are used to signal that you'd like exclusive control of the Leap. Other apps that provide OS-level interaction should be suppressed when you are sending heartbeats.

### Sending Heartbeats

Clients heartbeat by sending `{heartbeat: true}`. Heartbeats must be sent <100 ms from each other.

*Note:* the heartbeat mechanism has been replaced by "focused" messages as of protocol version #4. Only use heartbeat messages with versions 2 and 3 of the protocol.

## Version 3

### Changes

This version introduced server-side events. Events are structured in the following way.

    {event: {type: "deviceConnect", state: true}}

These were introduced to allow reliable reporting of events from the server. Currently only `deviceConnect` events are supported.

## Version 4

### Changes

This version removes heartbeating. Instead, applications request focus by sending focus events.

### Sending focus events

To receive frame messages containing tracking data, your application must take the Leap Motion control focus by sending the message: `{focused: true}` to the WebSocket server. 

When your application does not need frames -- for example, when your application moves to the background of the user's desktop -- send a `{focused: false}` message to the WebSocket server. Sending this message prevents the user from accidentally making changes in your application with Leap Motion interaction intended for another app. It can also reduce the CPU footprint of your app when it is in the background. To resume processing frames of tracking data, send another `{focused: true}` message.

If another application takes the Leap Motion control focus, your application loses focus automatically and won't receive frame messages even if you don't send a `{focused: false}` message.

### Background mode

In order to always receive frames, you must enable background mode. This can be achieved by sending `{background: true}`. By default, applications will not receive frames when they lose focus.
